//------------------------------------------------------ 2021-07-11 --------------
// Laminator temperature controler  (Arduino 1.8.13)
//
// Board: Arduino Nano Atmega328 16MHz
// Library:
//   TM1637 by Avishay Orpaz V1.2.0
//----------------------------- written by Hiroyuki Iizuka ---- Japan ------------
#include <TM1637Display.h>
#include <EEPROM.h>
#include <Encoder.h>
#include <MsTimer2.h>

//----------------------------------------------------
#define SERIES_RESISTANCE  4609.5 // ohm
//----------------------------------------------------

#define PERIOD    2000   // heater PWM period [msec]

#define TM1637_CLK    A1    // 7segment LED clock
#define TM1637_DIO    A0    // 7segment LED data

#define THERMISTOR         A2   // NTC thermistor

#define HEATER             A5   // photo triac
#define HEATER_OFF 0
#define HEATER_ON  1

#define ENCODER_A          9   // rotary encoder
#define ENCODER_B         10   // rotary encoder

#define LED_A             4   // LED A
#define LED_B             5   // LED B
#define LED_C             6   // LED C
#define LED_ON  0
#define LED_OFF 1

#define SLIDE_SW1         8   //   SW1=HIGH SW2=LOW   Temperature A 
#define SLIDE_SW2         7   //   SW1=HIGH SW2=HIGH  Temperature B
                              //   SW1=LOW  SW2=HIGH  Temperature C

#define MIN_TEMPERATURE  80
#define MAX_TEMPERATURE  150

#define MODE_A 1
#define MODE_B 2
#define MODE_C 3

#define EEPADDR_MODE_A_TEMPERATURE 0
#define EEPADDR_MODE_B_TEMPERATURE 1
#define EEPADDR_MODE_C_TEMPERATURE 2

TM1637Display display(TM1637_CLK, TM1637_DIO);
Encoder myEncoder(ENCODER_A, ENCODER_B);

long          target_temperature;     // target temperature
double        current_temperature;    // current temperature 
boolean       enable_sensor;
unsigned long disable_time;
unsigned char mode; 
boolean       mode_changed; 
unsigned char heater_pwm_duty;        // 0 to 100

double calc_Bconstant( 
  double Tn, // temperature celsius
  double Rn, // NTC thermistor resistance 
  double R0) // NTC thermistor resistance @ 25 degrees_celsius
{
  return log(Rn/R0) / (1/celsius_to_kelvin(Tn) - 1/celsius_to_kelvin(25.0));
}

//--------------- calc temperature Mitsubish material ----------------------
//  https://www.mmc.co.jp/adv/dev/japan/document/thermistor/thermistor02.html
double get_temperature_Mitsubishi(
  double series_resistance,
  double R0, // NTC thermistor resistance @ 25 degrees_celsius
  unsigned int B_constant, // @25℃
  double t1, // temperature celsius
  double t2, // temperature celsiusca
  double t3, // temperature celsius
  double R1, // NTC thermistor resistance @ t1 degrees celsius
  double R2, // NTC thermistor resistance @ t2 degrees celsius
  double R3, // NTC thermistor resistance @ t3 degrees celsius
  double E_calibration)  
{
  double Bn,Bc1,Bc2,Bc3;
  double Tprev,Tn,T1,T2,T3;
  double C,D,E;
  
  Bc1 = calc_Bconstant( t1, R1, R0 );
  Bc2 = calc_Bconstant( t2, R2, R0 );
  Bc3 = calc_Bconstant( t3, R3, R0 );
  T1 = celsius_to_kelvin( t1 );
  T2 = celsius_to_kelvin( t2 );
  T3 = celsius_to_kelvin( t3 );
  C = ((Bc1-Bc2)*(T2-T3)-(Bc2-Bc3)*(T1-T2)) / ((T1-T2)*(T2-T3)*(T1-T3));
  D = (Bc1-Bc2-C*(T1+T2)*(T1-T2)) / (T1-T2);
  E = Bc1 - D*T1 - C*T1*T1 + E_calibration;

  Tprev = get_temperature( series_resistance, R0, B_constant, true );
  for( unsigned char i = 0 ; i < 2 ; i++ ) {
    Bn = C * Tprev*Tprev + D * Tprev + E;
    Tn = get_temperature( series_resistance, R0, Bn, false );
    if( (Tprev - 0.01 < Tn) && (Tn < Tprev + 0.01) ) {
      break;
    } else {
      Tprev = Tn;
    }
  }

  return Tn;   // Kelvin
}
  
//----------------------- temperature conversion -----------------------------
double celsius_to_kelvin( double celsius )
{
  return celsius + 273.15;
}

double kelvin_to_celsius( double kelvin )
{
  return kelvin - 273.15;
}

//----------- calc temperature -----------------------------
double get_temperature(
  unsigned int series_resistance,
  double       R0, // NTC thermistor resistance @ 25 degrees_celsius
  unsigned int B_constant,
  boolean      exec_ad )
{
  static double ntc_resistance;

  if( exec_ad ) {
    ntc_resistance = get_ntc_resistance( series_resistance );
  }
  
  return 1 / (log(ntc_resistance/R0)/B_constant + 1/celsius_to_kelvin(25.0));
}  

//---------- A/D ntc resistance ---------------
double get_ntc_resistance( unsigned int series_resistance )
{
  double ad_value;

  ad_value = 0.0;
  for( unsigned char i = 0 ; i < 100 ; i++ ) {
    ad_value += (double)analogRead( THERMISTOR );
    delay(1);
  }
  ad_value /= 100;
 
  return ( ad_value * series_resistance ) / ( 1024 - ad_value );
}

//------- read mode slide switch ------------------------- 
unsigned char get_mode( void )
{
  static unsigned char prev_mode = MODE_A;
  unsigned char mode;
  
  if( digitalRead( SLIDE_SW1 ) == HIGH && digitalRead( SLIDE_SW2 ) == LOW  ) { mode = MODE_A; }
  if( digitalRead( SLIDE_SW1 ) == HIGH && digitalRead( SLIDE_SW2 ) == HIGH ) { mode = MODE_B; }
  if( digitalRead( SLIDE_SW1 ) == LOW  && digitalRead( SLIDE_SW2 ) == HIGH ) { mode = MODE_C; }  
  if( mode != prev_mode ) {
    prev_mode = mode;
    mode_changed = true;
  }

  return mode;
}

//------- EEPROM -> target_temperature  ----------------------
unsigned char get_target_temperature( unsigned char mode )
{
  unsigned char temperature;
  
  switch( mode ) {
   case MODE_A: temperature = EEPROM.read( EEPADDR_MODE_A_TEMPERATURE ); break;
   case MODE_B: temperature = EEPROM.read( EEPADDR_MODE_B_TEMPERATURE ); break;
   case MODE_C: temperature = EEPROM.read( EEPADDR_MODE_C_TEMPERATURE ); break;
  }

  if( (temperature < MIN_TEMPERATURE) || (MAX_TEMPERATURE < temperature) ) {
    temperature = MIN_TEMPERATURE;
    write_setpoint_temperature( temperature, mode );
  }

  return temperature;
}

//------- target_temperature -> EEPROM ----------------------
void write_setpoint_temperature( unsigned char temperature, unsigned char mode )
{
  switch( mode ) {
   case MODE_A: EEPROM.write( EEPADDR_MODE_A_TEMPERATURE, temperature ); break;
   case MODE_B: EEPROM.write( EEPADDR_MODE_B_TEMPERATURE, temperature ); break;
   case MODE_C: EEPROM.write( EEPADDR_MODE_C_TEMPERATURE, temperature ); break;
  }
}

//---------- MsTimer2 interrupt routine --------------
void heater()
{
  static unsigned char counter = 0;

  if( 100 <= counter ) {
    counter = 0;
  }
  if( counter < heater_pwm_duty ) {
    digitalWrite( HEATER, HEATER_ON  );
    switch( mode ) {
      case MODE_A:  digitalWrite( LED_A, LED_ON  ); digitalWrite( LED_B, LED_OFF ); digitalWrite( LED_C, LED_OFF ); break;
      case MODE_B:  digitalWrite( LED_A, LED_OFF ); digitalWrite( LED_B, LED_ON  ); digitalWrite( LED_C, LED_OFF ); break;
      case MODE_C:  digitalWrite( LED_A, LED_OFF ); digitalWrite( LED_B, LED_OFF ); digitalWrite( LED_C, LED_ON  ); break;
    } 
  } else {     
    digitalWrite( HEATER, HEATER_OFF );
    digitalWrite( LED_A, LED_OFF );
    digitalWrite( LED_B, LED_OFF );
    digitalWrite( LED_C, LED_OFF );
  }
  counter++;
}

//----------------------- setup -----------------------------------
void setup(){
  pinMode( THERMISTOR, INPUT );
  pinMode( ENCODER_A,  INPUT );
  pinMode( ENCODER_B,  INPUT );
  pinMode( SLIDE_SW1,  INPUT );
  pinMode( SLIDE_SW2,  INPUT );
  pinMode( LED_A,      OUTPUT );
  pinMode( LED_B,      OUTPUT );
  pinMode( LED_C,      OUTPUT );
  pinMode( HEATER,     OUTPUT );

  digitalWrite( HEATER, HEATER_OFF );
  digitalWrite( LED_A, LED_OFF );
  digitalWrite( LED_B, LED_OFF );
  digitalWrite( LED_C, LED_OFF );

  display.setBrightness(7);
  display.clear();

  mode = get_mode();
  mode_changed = false;
  target_temperature = get_target_temperature( mode );
  myEncoder.write( target_temperature * 4);
  display.showNumberDecEx(target_temperature * 10, 0b00100000, false);

  delay(1000);
  
  Serial.begin(9600);
  Serial.println("-------- Laminator temperature controller -------------------");
  
  enable_sensor = true;
  MsTimer2::set(PERIOD/100, heater);  
  MsTimer2::start();

  heater_pwm_duty = 0;
}

//----------------------- loop ------------------------------------
void loop()
{
  long encoder_value;

  if( enable_sensor ) {
    current_temperature = kelvin_to_celsius(get_temperature_Mitsubishi( SERIES_RESISTANCE, 100000, 4100, 50, 100, 150, 34470, 6092, 1569, -182 ));
    display.showNumberDecEx( current_temperature * 10, 0b00100000, false);
    enable_sensor = false;
    disable_time = millis();
    Serial.print("temp=");    Serial.print( current_temperature ); Serial.print(" ");
    Serial.print("target=");  Serial.print( target_temperature );  Serial.print(" ");
    Serial.print("PWM duty=");Serial.print( heater_pwm_duty );     Serial.println(" ");

    heater_pwm_duty = max( 0, min( ( target_temperature - current_temperature ) * 10, 100 ));
  }

  if( disable_time + 1000 < millis() ) {
    enable_sensor = true;  
  }
  
  encoder_value = myEncoder.read() / 4;
  if( encoder_value != target_temperature ) {
    if( encoder_value < MIN_TEMPERATURE ) {
      myEncoder.write( MIN_TEMPERATURE * 4 ); 
    } else if( MAX_TEMPERATURE < encoder_value ){
      myEncoder.write( MAX_TEMPERATURE * 4 ); 
    } else {
      target_temperature = encoder_value;
    }
    display.showNumberDecEx( target_temperature * 10, 0b00100000, false);
    write_setpoint_temperature( target_temperature, mode );
    enable_sensor = false;
    disable_time = millis();
  }

  mode = get_mode();
  if( mode_changed ) {
    target_temperature = get_target_temperature( mode );
    myEncoder.write( target_temperature * 4 );
    for( unsigned char i = 0 ; i < 3 ; i++ ) { // brink target temperature
      display.clear();
      delay(50);
      display.showNumberDecEx(target_temperature * 10, 0b00100000, false);
      delay(300);
    }
    mode_changed = false;
  }
}

//--------------- E N D -------------------------------------------------------------------
